/*
 * ffsox_clone.c
 * Copyright (C) 2011, 2012 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef NDEBUG /* N.B. assert used with active statements so enable always. */
#undef NDEBUG /* Must undef above assert.h or other that might include it. */
#endif

#include <assert.h>
#include <math.h>
#include "ffsox.h"

#ifndef WIN32
  #define stricmp strcasecmp
  #define strnicmp strncasecmp
#endif

///////////////////////////////////////////////////////////////////////////////
typedef struct ffo_st {
  int skip;
  AVStream *st;
  AVCodec *enc;
  AVCodecContext *ec;
  int index;
  int64_t idts;
} ffo_st_t;

static ffo_st_t *ffo_st_clone(ffo_st_t *ost, ffi_st_t *ist,
    AVFormatContext *oc);
static ffo_st_t *ffo_st_cleanup(ffo_st_t *ost);

///////////////////////////////////////////////////////////////////////////////
static ffo_st_t *ffo_st_clone(ffo_st_t *ost, ffi_st_t *ist,
    AVFormatContext *oc)
{
  AVCodecContext *dc=ist->dc;
  uint64_t extra_size=(uint64_t)dc->extradata_size
      +FF_INPUT_BUFFER_PADDING_SIZE;
  AVStream *st;
  AVCodecContext *ec;
  int copy_tb=-1;

  if (extra_size>INT_MAX)
    goto error;

  // from "new_*_stream()"
  if (NULL==(ost->st=st=avformat_new_stream(oc,NULL)))
    goto error;

  ost->skip=0;
  ost->index=st->index;
  ost->ec=ec=st->codec;
  //avcodec_get_context_defaults3(ec, ost->enc);
  ec->codec_type=dc->codec_type;

  if (oc->oformat->flags&AVFMT_GLOBALHEADER)
    ec->flags|=CODEC_FLAG_GLOBAL_HEADER;

  ost->idts=0;

  // from "transcode()"
  // if stream_copy is selected, no need to decode or encode.
  ec->codec_id=dc->codec_id;
  ec->codec_type=dc->codec_type;

  if (!ec->codec_tag) {
    if (!oc->oformat->codec_tag
        ||av_codec_get_id (oc->oformat->codec_tag,dc->codec_tag)
            ==ec->codec_id
        ||av_codec_get_tag(oc->oformat->codec_tag,dc->codec_id)<=0)

    ec->codec_tag = dc->codec_tag;
  }

  ec->bit_rate=dc->bit_rate;
  ec->rc_max_rate=dc->rc_max_rate;
  ec->rc_buffer_size=dc->rc_buffer_size;
  ec->field_order=dc->field_order;
  
  if (NULL==(ec->extradata=av_mallocz(extra_size)))
    goto error;

  memcpy(ec->extradata, dc->extradata, dc->extradata_size);
  ec->extradata_size=dc->extradata_size;

  ec->time_base=ist->st->time_base;
  /*
   * Avi is a special case here because it supports variable fps but
   * having the fps and timebase differe significantly adds quite some
   * overhead
   */
  if(!strcmp(oc->oformat->name, "avi")) {
    if (   copy_tb<0 && av_q2d(dc->time_base)*dc->ticks_per_frame > 2*av_q2d(ist->st->time_base)
                     && av_q2d(ist->st->time_base) < 1.0/500
        || copy_tb==0){
        ec->time_base = dc->time_base;
        ec->time_base.num *= dc->ticks_per_frame;
        ec->time_base.den *= 2;
    }
  } else if(!(oc->oformat->flags & AVFMT_VARIABLE_FPS)
          && strcmp(oc->oformat->name, "mov") && strcmp(oc->oformat->name, "mp4") && strcmp(oc->oformat->name, "3gp")
          && strcmp(oc->oformat->name, "3g2") && strcmp(oc->oformat->name, "psp") && strcmp(oc->oformat->name, "ipod")
  ) {
    if(   copy_tb<0 && av_q2d(dc->time_base)*dc->ticks_per_frame > av_q2d(ist->st->time_base)
                    && av_q2d(ist->st->time_base) < 1.0/500
       || copy_tb==0){
        ec->time_base = dc->time_base;
        ec->time_base.num *= dc->ticks_per_frame;
    }
  }

  av_reduce(&ec->time_base.num, &ec->time_base.den,
      ec->time_base.num, ec->time_base.den, INT_MAX);

  switch (ec->codec_type) {
    case AVMEDIA_TYPE_AUDIO:
      ec->channel_layout=dc->channel_layout;
      ec->sample_rate=dc->sample_rate;
      ec->channels=dc->channels;
      ec->frame_size=dc->frame_size;
      ec->audio_service_type=dc->audio_service_type;
      ec->block_align=dc->block_align;

      if (ec->codec_id==CODEC_ID_MP3&&ec->block_align==1)
        ec->block_align=0;
      else if (ec->codec_id==CODEC_ID_AC3)
        ec->block_align=0;

      break;

    case AVMEDIA_TYPE_VIDEO:
      ec->sample_aspect_ratio=st->sample_aspect_ratio=dc->sample_aspect_ratio;
      ec->pix_fmt=dc->pix_fmt;
      ec->width=dc->width;
      ec->height=dc->height;
      ec->has_b_frames=dc->has_b_frames;
      break;

    case AVMEDIA_TYPE_SUBTITLE:
      ec->width=dc->width;
      ec->height=dc->height;
      break;

    default:
      //goto error;
      break;
  }

  return ost;
error:
  return NULL;
}

static ffo_st_t *ffo_st_cleanup(ffo_st_t *ost)
{
  return ost;
}

///////////////////////////////////////////////////////////////////////////////
typedef struct ffo {
  ffi_t *ffi;
  AVFormatContext *oc;
  ffo_st_t *st;
  ffo_st_t *ast;
} ffo_t;

static ffo_t *ffo_init(ffo_t *ffo, ffi_t *ffi, pb_str_t *fn, char **meta);
static ffo_t *ffo_cleanup(ffo_t *ffo);

static int ffo_loop(ffo_t *ffo);

///////////////////////////////////////////////////////////////////////////////
static void ffo_clone_dict_filtered(AVDictionary **od, AVDictionary *id,
    char **meta)
{
  AVDictionaryEntry *mtag=NULL;
  char **m;

  while (NULL!=(mtag=av_dict_get(id,"",mtag,AV_DICT_IGNORE_SUFFIX))) {
    for (m=meta;m&&*m;++m)
      if (!stricmp(*m,mtag->key))
        goto next_mtag;

    av_dict_set(od,mtag->key,mtag->value,0);
  next_mtag:
    continue;
  }
}

static void ffo_add_dict(AVDictionary **od, char **meta)
{
  char **m;

  for (m=meta;m&&*m;++m) {
    char *val=*m+strlen(*m)+1;

    if (*val)
      av_dict_set(od,*m,val,0);
  }
}

static ffo_t *ffo_init(ffo_t *ffo, ffi_t *ffi, pb_str_t *fn, char **meta)
{
  AVOutputFormat *oformat;
  AVFormatContext *oc;
  int i;

  memset(ffo, 0, sizeof *ffo);
  ffo->ffi=ffi;

  if (NULL==(oformat=av_guess_format(NULL,PS_UTF8(fn),NULL)))
    goto error;

  if (NULL==(ffo->oc=oc=avformat_alloc_context()))
    goto error;

  oc->oformat=oformat;
  //av_strlcpy(oc->filename, fn, sizeof(oc->filename));
  strncpy(oc->filename,PS_UTF8(fn),sizeof(oc->filename));

  if (NULL==(ffo->st=calloc(ffi->nb_streams, sizeof *ffo->st)))
    goto error;

  for (i=0;i<ffi->nb_streams;++i) {
    ffi_st_t *ist=ffi->st+i;
    ffo_st_t *ost=ffo->st+i;

    if (NULL==ist->dc
	    ||(AVMEDIA_TYPE_VIDEO==ist->dc->codec_type
        &&AV_CODEC_ID_NONE==oc->oformat->video_codec)) {

      ost->skip=1;
      continue;
    }
    else if (NULL==ffo_st_clone(ffo->st+i,ist,oc))
      goto error;

    if (ffi->ast==ffi->st+i)
      ffo->ast=ffo->st+i;
  }

  ffo_clone_dict_filtered(&oc->metadata,ffi->ic->metadata,meta);

#if defined (FFO_CLONE_STREAM_META) // {
  for (i=0;i<ffi->nb_streams;++i) {
    ffi_st_t *ist=ffi->st+i;
    ffo_st_t *ost=ffo->st+i;

    ffo_clone_dict_filtered(&oc->metadata,ist->st->metadata,meta);
  }
#endif // }

  ffo_add_dict(&oc->metadata,meta);

  if (0!=(oformat->flags&AVFMT_NOFILE)
      ||avio_open(&oc->pb,PS_UTF8(fn),AVIO_FLAG_WRITE)<0)
    goto error;

  return ffo;
error:
#if defined (WIN32)
  _ftprintf(stderr,_T("Error opening \"%s\".\n"),PS_TCS(fn));
#else
  fprintf(stderr,"Error opening \"%s\".\n",PS_UTF8(fn));
#endif
  ffo_cleanup(ffo);

  return NULL;
}

static ffo_t *ffo_cleanup(ffo_t *ffo)
{
  AVFormatContext *oc=ffo->oc;
  int i;

  if (NULL!=oc&&NULL!=oc->pb) {
    avio_close(oc->pb);
    oc->pb=NULL;
  }

  if (NULL!=ffo->st) {
    free(ffo->st);
    ffo->st=NULL;
  }

  return ffo;
}

#if defined (__ORIGINAL_FFMPEG__)
static void write_frame(AVFormatContext *s, AVPacket *pkt, OutputStream *ost)
#else
static int ffo_write_frame(AVFormatContext *s, AVPacket *pkt, ffi_st_t *ist, ffo_st_t *ost)
#endif
{
#if defined (__ORIGINAL_FFMPEG__)
    AVBitStreamFilterContext *bsfc = ost->bitstream_filters;
#endif
    AVCodecContext          *avctx = ost->st->codec;
    int ret;

#if defined (__ORIGINAL_FFMPEG__)
    if ((avctx->codec_type == AVMEDIA_TYPE_VIDEO && video_sync_method == VSYNC_DROP) ||
        (avctx->codec_type == AVMEDIA_TYPE_AUDIO && audio_sync_method < 0))
        pkt->pts = pkt->dts = AV_NOPTS_VALUE;
#endif

    if ((avctx->codec_type == AVMEDIA_TYPE_AUDIO || avctx->codec_type == AVMEDIA_TYPE_VIDEO) && pkt->dts != AV_NOPTS_VALUE) {
        int64_t max = ost->st->cur_dts + !(s->oformat->flags & AVFMT_TS_NONSTRICT);
        if (ost->st->cur_dts && ost->st->cur_dts != AV_NOPTS_VALUE &&  max > pkt->dts) {
            av_log(s, max - pkt->dts > 2 || avctx->codec_type == AVMEDIA_TYPE_VIDEO ? AV_LOG_WARNING : AV_LOG_DEBUG,
                   "st:%d PTS: %"PRId64" DTS: %"PRId64" < %"PRId64" invalid, clipping\n", pkt->stream_index, pkt->pts, pkt->dts, max);
            if(pkt->pts >= pkt->dts)
                pkt->pts = FFMAX(pkt->pts, max);
            pkt->dts = max;
        }
    }

#if defined (__ORIGINAL_FFMPEG__)
    /*
     * Audio encoders may split the packets --  #frames in != #packets out.
     * But there is no reordering, so we can limit the number of output packets
     * by simply dropping them here.
     * Counting encoded video frames needs to be done separately because of
     * reordering, see do_video_out()
     */
    if (!(avctx->codec_type == AVMEDIA_TYPE_VIDEO && avctx->codec)) {
        if (ost->frame_number >= ost->max_frames) {
            av_free_packet(pkt);
            return;
        }
        ost->frame_number++;
    }

    while (bsfc) {
        AVPacket new_pkt = *pkt;
        int a = av_bitstream_filter_filter(bsfc, avctx, NULL,
                                           &new_pkt.data, &new_pkt.size,
                                           pkt->data, pkt->size,
                                           pkt->flags & AV_PKT_FLAG_KEY);
        if(a == 0 && new_pkt.data != pkt->data && new_pkt.destruct) {
            uint8_t *t = av_malloc(new_pkt.size + FF_INPUT_BUFFER_PADDING_SIZE); //the new should be a subset of the old so cannot overflow
            if(t) {
                memcpy(t, new_pkt.data, new_pkt.size);
                memset(t + new_pkt.size, 0, FF_INPUT_BUFFER_PADDING_SIZE);
                new_pkt.data = t;
                a = 1;
            } else
                a = AVERROR(ENOMEM);
        }
        if (a > 0) {
            av_free_packet(pkt);
            new_pkt.destruct = av_destruct_packet;
        } else if (a < 0) {
            av_log(NULL, AV_LOG_ERROR, "Failed to open bitstream filter %s for stream %d with codec %s",
                   bsfc->filter->name, pkt->stream_index,
                   avctx->codec ? avctx->codec->name : "copy");
            print_error("", a);
            if (exit_on_error)
                exit_program(1);
        }
        *pkt = new_pkt;

        bsfc = bsfc->next;
    }
#endif

    pkt->stream_index = ost->index;
    ret = av_interleaved_write_frame(s, pkt);
#if defined (__ORIGINAL_FFMPEG__)
    if (ret < 0) {
        print_error("av_interleaved_write_frame()", ret);
        exit_program(1);
    }
#else
    return ret;
#endif
}

#if 0 // {
#if defined (__ORIGINAL_FFMPEG__)
static void do_streamcopy(InputStream *ist, OutputStream *ost, const AVPacket *pkt)
#else
static int ffo_streamcopy(ffo_t *ffo, const AVPacket *pkt)
#endif
{
#if defined (__ORIGINAL_FFMPEG__)
    OutputFile *of = output_files[ost->file_index];
    int64_t ost_tb_start_time = av_rescale_q(of->start_time, AV_TIME_BASE_Q, ost->st->time_base);
    AVPicture pict;
#else
    int code=-1;
    ffi_t *ffi=ffo->ffi;
    ffi_st_t *ist=ffi->st+pkt->stream_index;
    ffo_st_t *ost=ffo->st+pkt->stream_index;
    int64_t ost_tb_start_time = 0;
    AVPicture pict;
#endif
    AVPacket opkt;

    av_init_packet(&opkt);

#if defined (__ORIGINAL_FFMPEG__)
    if ((!ost->frame_number && !(pkt->flags & AV_PKT_FLAG_KEY)) &&
        !ost->copy_initial_nonkeyframes)
        return;

    if (of->recording_time != INT64_MAX &&
        ist->pts >= of->recording_time + of->start_time) {
        close_output_stream(ost);
        return;
    }

    /* force the input stream PTS */
    if (ost->st->codec->codec_type == AVMEDIA_TYPE_AUDIO)
        audio_size += pkt->size;
    else if (ost->st->codec->codec_type == AVMEDIA_TYPE_VIDEO) {
        video_size += pkt->size;
        ost->sync_opts++;
    } else if (ost->st->codec->codec_type == AVMEDIA_TYPE_SUBTITLE) {
        subtitle_size += pkt->size;
    }
#endif

    if (pkt->pts != AV_NOPTS_VALUE)
        opkt.pts = av_rescale_q(pkt->pts, ist->st->time_base, ost->st->time_base) - ost_tb_start_time;
    else
        opkt.pts = AV_NOPTS_VALUE;

    if (pkt->dts == AV_NOPTS_VALUE)
        opkt.dts = av_rescale_q(ist->dts, AV_TIME_BASE_Q, ost->st->time_base);
    else
        opkt.dts = av_rescale_q(pkt->dts, ist->st->time_base, ost->st->time_base);
    opkt.dts -= ost_tb_start_time;

    opkt.duration = av_rescale_q(pkt->duration, ist->st->time_base, ost->st->time_base);
    opkt.flags    = pkt->flags;

    // FIXME remove the following 2 lines they shall be replaced by the bitstream filters
    if (  ost->st->codec->codec_id != AV_CODEC_ID_H264
       && ost->st->codec->codec_id != AV_CODEC_ID_MPEG1VIDEO
       && ost->st->codec->codec_id != AV_CODEC_ID_MPEG2VIDEO
       && ost->st->codec->codec_id != AV_CODEC_ID_VC1
       ) {
        if (av_parser_change(ist->st->parser, ost->st->codec, &opkt.data, &opkt.size, pkt->data, pkt->size, pkt->flags & AV_PKT_FLAG_KEY))
            opkt.destruct = av_destruct_packet;
    } else {
        opkt.data = pkt->data;
        opkt.size = pkt->size;
    }

#if defined (__ORIGINAL_FFMPEG__)
    if (ost->st->codec->codec_type == AVMEDIA_TYPE_VIDEO && (of->ctx->oformat->flags & AVFMT_RAWPICTURE)) {
        /* store AVPicture in AVPacket, as expected by the output format */
        avpicture_fill(&pict, opkt.data, ost->st->codec->pix_fmt, ost->st->codec->width, ost->st->codec->height);
        opkt.data = (uint8_t *)&pict;
        opkt.size = sizeof(AVPicture);
        opkt.flags |= AV_PKT_FLAG_KEY;
    }

    write_frame(of->ctx, &opkt, ost);
    ost->st->codec->frame_number++;
    av_free_packet(&opkt);
#else
    if (ost->st->codec->codec_type == AVMEDIA_TYPE_VIDEO && (ffo->oc->oformat->flags & AVFMT_RAWPICTURE)) {
        /* store AVPicture in AVPacket, as expected by the output format */
        avpicture_fill(&pict, opkt.data, ost->st->codec->pix_fmt, ost->st->codec->width, ost->st->codec->height);
        opkt.data = (uint8_t *)&pict;
        opkt.size = sizeof(AVPicture);
        opkt.flags |= AV_PKT_FLAG_KEY;
    }

    if (0!=ffo_write_frame(ffo->oc,&opkt,ist,ost))
      goto cleanup;

    ost->st->codec->frame_number++;
    code=0;
cleanup:
    av_free_packet(&opkt);

    return code;
#endif
}
#else // } {
#if defined (__ORIGINAL_FFMPEG__)
static void do_streamcopy(InputStream *ist, OutputStream *ost, const AVPacket *pkt)
#else
static int ffo_streamcopy(ffo_t *ffo, const AVPacket *pkt)
#endif
{
#if defined (__ORIGINAL_FFMPEG__) // {
    OutputFile *of = output_files[ost->file_index];
    int64_t ost_tb_start_time = av_rescale_q(of->start_time, AV_TIME_BASE_Q, ost->st->time_base);
#else // } {
    int code=-1;
    ffi_t *ffi=ffo->ffi;
    ffi_st_t *ist=ffi->st+pkt->stream_index;
    ffo_st_t *ost=ffo->st+pkt->stream_index;
    int64_t ost_tb_start_time = 0;
#endif // }
    AVPicture pict;
    AVPacket opkt;

    av_init_packet(&opkt);

#if defined (__ORIGINAL_FFMPEG__) // {
    if ((!ost->frame_number && !(pkt->flags & AV_PKT_FLAG_KEY)) &&
        !ost->copy_initial_nonkeyframes)
        return;

    if (!ost->frame_number && ist->pts < of->start_time &&
        !ost->copy_prior_start)
        return;

    if (of->recording_time != INT64_MAX &&
        ist->pts >= of->recording_time + of->start_time) {
        close_output_stream(ost);
        return;
    }

    /* force the input stream PTS */
    if (ost->st->codec->codec_type == AVMEDIA_TYPE_AUDIO)
        audio_size += pkt->size;
    else if (ost->st->codec->codec_type == AVMEDIA_TYPE_VIDEO) {
        video_size += pkt->size;
        ost->sync_opts++;
    } else if (ost->st->codec->codec_type == AVMEDIA_TYPE_SUBTITLE) {
        subtitle_size += pkt->size;
    }
#endif // }

    if (pkt->pts != AV_NOPTS_VALUE)
        opkt.pts = av_rescale_q(pkt->pts, ist->st->time_base, ost->st->time_base) - ost_tb_start_time;
    else
        opkt.pts = AV_NOPTS_VALUE;

    if (pkt->dts == AV_NOPTS_VALUE)
        opkt.dts = av_rescale_q(ist->dts, AV_TIME_BASE_Q, ost->st->time_base);
    else
        opkt.dts = av_rescale_q(pkt->dts, ist->st->time_base, ost->st->time_base);
    opkt.dts -= ost_tb_start_time;

#if defined (__ORIGINAL_FFMPEG__) // {
    if (ost->st->codec->codec_type == AVMEDIA_TYPE_AUDIO && pkt->dts != AV_NOPTS_VALUE) {
        int duration = av_get_audio_frame_duration(ist->st->codec, pkt->size);
        if(!duration)
            duration = ist->st->codec->frame_size;
        opkt.dts = opkt.pts = av_rescale_delta(ist->st->time_base, pkt->dts,
                                               (AVRational){1, ist->st->codec->sample_rate}, duration, &ist->filter_in_rescale_delta_last,
                                               ost->st->time_base) - ost_tb_start_time;
    }
#endif // }

    opkt.duration = av_rescale_q(pkt->duration, ist->st->time_base, ost->st->time_base);
    opkt.flags    = pkt->flags;

    // FIXME remove the following 2 lines they shall be replaced by the bitstream filters
    if (  ost->st->codec->codec_id != AV_CODEC_ID_H264
       && ost->st->codec->codec_id != AV_CODEC_ID_MPEG1VIDEO
       && ost->st->codec->codec_id != AV_CODEC_ID_MPEG2VIDEO
       && ost->st->codec->codec_id != AV_CODEC_ID_VC1
       ) {
        if (av_parser_change(ist->st->parser, ost->st->codec, &opkt.data, &opkt.size, pkt->data, pkt->size, pkt->flags & AV_PKT_FLAG_KEY)) {
            opkt.buf = av_buffer_create(opkt.data, opkt.size, av_buffer_default_free, NULL, 0);
            if (!opkt.buf)
                exit(1);
        }
    } else {
        opkt.data = pkt->data;
        opkt.size = pkt->size;
    }

#if defined (__ORIGINAL_FFMPEG__) // {
    if (ost->st->codec->codec_type == AVMEDIA_TYPE_VIDEO && (of->ctx->oformat->flags & AVFMT_RAWPICTURE)) {
        /* store AVPicture in AVPacket, as expected by the output format */
        avpicture_fill(&pict, opkt.data, ost->st->codec->pix_fmt, ost->st->codec->width, ost->st->codec->height);
        opkt.data = (uint8_t *)&pict;
        opkt.size = sizeof(AVPicture);
        opkt.flags |= AV_PKT_FLAG_KEY;
    }

    write_frame(of->ctx, &opkt, ost);
    ost->st->codec->frame_number++;
#else // } {
    if (ost->st->codec->codec_type == AVMEDIA_TYPE_VIDEO && (ffo->oc->oformat->flags & AVFMT_RAWPICTURE)) {
        /* store AVPicture in AVPacket, as expected by the output format */
        avpicture_fill(&pict, opkt.data, ost->st->codec->pix_fmt, ost->st->codec->width, ost->st->codec->height);
        opkt.data = (uint8_t *)&pict;
        opkt.size = sizeof(AVPicture);
        opkt.flags |= AV_PKT_FLAG_KEY;
    }

    if (0!=ffo_write_frame(ffo->oc,&opkt,ist,ost))
      goto cleanup;

    ost->st->codec->frame_number++;
    code=0;
cleanup:
    av_free_packet(&opkt);

    return code;
#endif // }
}
#endif // }

/* pkt = NULL means EOF (needed to flush decoder buffers) */
#if defined (__ORIGINAL_FFMPEG__)
static int output_packet(InputStream *ist, const AVPacket *pkt)
#else
static int ffo_output_packet(ffo_t *ffo, const AVPacket *pkt)
#endif
{
#if ! defined (__ORIGINAL_FFMPEG__)
    ffi_t *ffi=ffo->ffi;
    ffi_st_t *ist=ffi->st+pkt->stream_index;
    //ffo_st_t *ost=ffo->st+pkt->stream_index;
#endif
    int ret = 0, i;
    int got_output;
    AVPacket avpkt;

    if (!ist->saw_first_ts) {
        ist->dts = ist->st->avg_frame_rate.num ? - ist->st->codec->has_b_frames * AV_TIME_BASE / av_q2d(ist->st->avg_frame_rate) : 0;
#if defined (__ORIGINAL_FFMPEG__)
        ist->pts = 0;
#endif
        if (pkt != NULL && pkt->pts != AV_NOPTS_VALUE && !ist->decoding_needed) {
            ist->dts += av_rescale_q(pkt->pts, ist->st->time_base, AV_TIME_BASE_Q);
#if defined (__ORIGINAL_FFMPEG__)
            ist->pts = ist->dts; //unused but better to set it to a value thats not totally wrong
#endif
        }
        ist->saw_first_ts = 1;
    }

    if (ist->next_dts == AV_NOPTS_VALUE)
        ist->next_dts = ist->dts;
#if defined (__ORIGINAL_FFMPEG__)
    if (ist->next_pts == AV_NOPTS_VALUE)
        ist->next_pts = ist->pts;
#endif

    if (pkt == NULL) {
        /* EOF handling */
        av_init_packet(&avpkt);
        avpkt.data = NULL;
        avpkt.size = 0;
#if defined (__ORIGINAL_FFMPEG__)
        goto handle_eof;
#endif
    } else {
        avpkt = *pkt;
    }

    if (pkt->dts != AV_NOPTS_VALUE) {
        ist->next_dts = ist->dts = av_rescale_q(pkt->dts, ist->st->time_base, AV_TIME_BASE_Q);
#if defined (__ORIGINAL_FFMPEG__)
        if (ist->st->codec->codec_type != AVMEDIA_TYPE_VIDEO || !ist->decoding_needed)
            ist->next_pts = ist->pts = av_rescale_q(pkt->dts, ist->st->time_base, AV_TIME_BASE_Q);
#endif
    }

    // while we have more to decode or while the decoder did output something on EOF
#if defined (__ORIGINAL_FFMPEG__)
    while (ist->decoding_needed && (avpkt.size > 0 || (!pkt && got_output))) {
        int duration;
    handle_eof:

#if defined (__ORIGINAL_FFMPEG__)
        ist->pts = ist->next_pts;
#endif
        ist->dts = ist->next_dts;

#if defined (__ORIGINAL_FFMPEG__)
        if (avpkt.size && avpkt.size != pkt->size) {
            av_log(NULL, ist->showed_multi_packet_warning ? AV_LOG_VERBOSE : AV_LOG_WARNING,
                   "Multiple frames in a packet from stream %d\n", pkt->stream_index);
            ist->showed_multi_packet_warning = 1;
        }
#endif

        switch (ist->st->codec->codec_type) {
        case AVMEDIA_TYPE_AUDIO:
            ret = decode_audio    (ist, &avpkt, &got_output);
            break;
        case AVMEDIA_TYPE_VIDEO:
            ret = decode_video    (ist, &avpkt, &got_output);
            if (avpkt.duration) {
                duration = av_rescale_q(avpkt.duration, ist->st->time_base, AV_TIME_BASE_Q);
            } else if(ist->st->codec->time_base.num != 0 && ist->st->codec->time_base.den != 0) {
                int ticks= ist->st->parser ? ist->st->parser->repeat_pict+1 : ist->st->codec->ticks_per_frame;
                duration = ((int64_t)AV_TIME_BASE *
                                ist->st->codec->time_base.num * ticks) /
                                ist->st->codec->time_base.den;
            } else
                duration = 0;

            if(ist->dts != AV_NOPTS_VALUE && duration) {
                ist->next_dts += duration;
            }else
                ist->next_dts = AV_NOPTS_VALUE;

#if defined (__ORIGINAL_FFMPEG__)
            if (got_output)
                ist->next_pts += duration; //FIXME the duration is not correct in some cases
#endif
            break;
        case AVMEDIA_TYPE_SUBTITLE:
            ret = transcode_subtitles(ist, &avpkt, &got_output);
            break;
        default:
            return -1;
        }

        if (ret < 0)
            return ret;

        avpkt.dts=
        avpkt.pts= AV_NOPTS_VALUE;

        // touch data and size only if not EOF
        if (pkt) {
            if(ist->st->codec->codec_type != AVMEDIA_TYPE_AUDIO)
                ret = avpkt.size;
            avpkt.data += ret;
            avpkt.size -= ret;
        }
        if (!got_output) {
            continue;
        }
    }
#endif

    /* handle stream copy */
    if (!ist->decoding_needed) {
#if defined (__ORIGINAL_FFMPEG__)
        rate_emu_sleep(ist);
#endif
        ist->dts = ist->next_dts;
        switch (ist->st->codec->codec_type) {
        case AVMEDIA_TYPE_AUDIO:
            ist->next_dts += ((int64_t)AV_TIME_BASE * ist->st->codec->frame_size) /
                             ist->st->codec->sample_rate;
            break;
        case AVMEDIA_TYPE_VIDEO:
            if (pkt->duration) {
                ist->next_dts += av_rescale_q(pkt->duration, ist->st->time_base, AV_TIME_BASE_Q);
            } else if(ist->st->codec->time_base.num != 0) {
                int ticks= ist->st->parser ? ist->st->parser->repeat_pict + 1 : ist->st->codec->ticks_per_frame;
                ist->next_dts += ((int64_t)AV_TIME_BASE *
                                  ist->st->codec->time_base.num * ticks) /
                                  ist->st->codec->time_base.den;
            }
            break;
        }
#if defined (__ORIGINAL_FFMPEG__)
        ist->pts = ist->dts;
        ist->next_pts = ist->next_dts;
#endif
    }
#if defined (__ORIGINAL_FFMPEG__)
    for (i = 0; pkt && i < nb_output_streams; i++) {
        OutputStream *ost = output_streams[i];

        if (!check_output_constraints(ist, ost) || ost->encoding_needed)
            continue;

        do_streamcopy(ist, ost, pkt);
    }
#else
    if (0!=(ret=ffo_streamcopy(ffo,pkt)))
        return ret;
#endif

    return 0;
}

static int ffo_loop(ffo_t *ffo)
{
  ffi_t *ffi=ffo->ffi;
  AVFormatContext *ic=ffi->ic;
  AVFormatContext *oc=ffo->oc;
  AVPacket pkt;
  int count=0;

  if (0!=avformat_write_header(oc,NULL))
    return -1;

  while (0==av_read_frame(ic, &pkt)) {
    ffi_st_t *ist=ffi->st+pkt.stream_index;
    ffo_st_t *ost=ffo->st+pkt.stream_index;
    int ret=0;

    if (0==ost->skip) {
      if (AV_NOPTS_VALUE!=pkt.dts) {
        pkt.dts+=av_rescale_q(ffi->ts_offset,AV_TIME_BASE_Q,
            ist->st->time_base);
      }

      if (AV_NOPTS_VALUE!=pkt.pts) {
        pkt.pts+=av_rescale_q(ffi->ts_offset,AV_TIME_BASE_Q,
            ist->st->time_base);
      }

      ret=ffo_output_packet(ffo,&pkt);
    }

    av_free_packet(&pkt);

    if (0!=ret)
      return ret;

    ++count;
  }

  av_write_trailer(oc);
  avio_flush(oc->pb);

  return 0==count?-1:0;
}

///////////////////////////////////////////////////////////////////////////////
int ffsox_clone(pb_str_t *fn1, pb_str_t *fn2, int mode, char **meta)
{
  int code=-1;
  ffi_t ffi;
  ffo_t ffo;

  memset(&ffi,0,sizeof ffi);
  memset(&ffo,0,sizeof ffo);

  if (NULL==ffi_init(&ffi,fn1,mode|FFI_STREAM_META|FFI_OPEN))
	goto cleanup;

  if (NULL==ffo_init(&ffo,&ffi,fn2,meta))
	goto cleanup;

  code=ffo_loop(&ffo);
cleanup:
  ffo_cleanup(&ffo);
  ffi_cleanup(&ffi);

  if (0!=code) {
#if defined (WIN32)
    _tunlink(fn2->tcs);
#else
    unlink(fn2->utf8);
#endif
  }

  return code;
}
